/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Wez Furlong <wez@php.net>                                    |
  | Portions influenced by libevent, written by                          |
  |             Niels Provos <provos@citi.umich.edu>                     |
  +----------------------------------------------------------------------+
*/

/* $Id: select.c,v 1.3 2004/12/16 03:30:18 wez Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_event.h"
#include "php_event_int.h"

#ifndef howmany
# define howmany(x, y)   (((x)+((y)-1))/(y))
#endif

#ifdef PHP_WIN32

/* win32 is good, bad and ugly.
 * The select based engine will only work with sockets.
 * To do stuff with files and other regular HANDLEs, we need real
 * overlapped IO and the win32 engine instead.
 */

# undef FD_SET
/* this version of FD_SET has faith that we always allocate a big enough set */
/* we can't simply call it FD_SET, because that is a typedef for fd_set in winsock *urgh* */
static void win_FD_SET(int fd, fd_set *set) {
	u_int i;

	for (i = 0; i < set->fd_count; i++) {
		if (set->fd_array[i] == fd)
			return;
	}
	set->fd_array[set->fd_count++] = fd;
}
#define FD_SET(f,s)	win_FD_SET(f,s)
#endif

/* select engine */

struct select_state {
	/* highest numbered fd, except on windows; where it is ignored */
	int max_fd;
	int fdsz;
	fd_set *rfds;
	fd_set *wfds;
	fd_set *efds;
};

static void *select_event_init(TSRMLS_D) {
	struct select_state *st = ecalloc(1, sizeof(*st));	
	st->max_fd = -1;
	return st;
}

static void select_event_fini(void *arg TSRMLS_DC)
{
	struct select_state *st = (struct select_state*)arg;
	efree(st->rfds);
	efree(st->wfds);
	efree(st->efds);
	efree(st);
}

static int select_event_add(void *arg, struct php_event *ev TSRMLS_DC)
{
	struct select_state *st = (struct select_state*)arg;

	if (st->max_fd < ev->ev_fd)
		st->max_fd = ev->ev_fd;

	return 0;
}

static int select_event_del(void *arg, struct php_event *ev TSRMLS_DC)
{
	struct select_state *st = (struct select_state*)arg;

	/* NOP */

	return 0;
}

static int select_event_recalc(void *arg, int max TSRMLS_DC)
{
	struct select_state *st = (struct select_state*)arg;
	struct php_event *ev;
	int fdsz;

#ifndef PHP_WIN32
	if (st->max_fd < max)
		st->max_fd = max;

	if (st->max_fd <= 0) {
		TAILQ_FOREACH(ev, &EVENT_G(eventqueue), ev_next) {
			if (ev->ev_fd > st->max_fd)
				st->max_fd = ev->ev_fd;
		}
	}

	fdsz = howmany(st->max_fd + 1, NFDBITS) * sizeof(fd_mask);
#else
	fdsz = sizeof(u_int) + (EVENT_G(event_count) * sizeof(HANDLE));
#endif

	if (fdsz > st->fdsz) {
		st->rfds = erealloc(st->rfds, fdsz);
		st->wfds = erealloc(st->wfds, fdsz);	
		st->efds = erealloc(st->efds, fdsz);
		st->fdsz = fdsz;
	}

	return 0;
}

static int select_event_dispatch(void *arg, struct timeval *tv TSRMLS_DC)
{
	struct select_state *st = (struct select_state*)arg;
	struct php_event *ev, *next;
	int maxfd, res;

#ifdef PHP_WIN32
	FD_ZERO(st->rfds);
	FD_ZERO(st->wfds);
	FD_ZERO(st->efds);
#else
	memset(st->rfds, 0, st->fdsz);
	memset(st->wfds, 0, st->fdsz);
	memset(st->efds, 0, st->fdsz);
#endif

	TAILQ_FOREACH(ev, &EVENT_G(eventqueue), ev_next) {
		if (ev->ev_events & EV_WRITE) {
			FD_SET(ev->ev_fd, st->wfds);
		}
		if (ev->ev_events & EV_READ) {
			FD_SET(ev->ev_fd, st->rfds);
		}
		if (ev->ev_events & EV_EXCEP) {
			FD_SET(ev->ev_fd, st->efds);
		}
	}

	if (st->max_fd == 0) {
		return -1;
	} else {
		res = select(st->max_fd + 1, st->rfds, st->wfds, st->efds, tv);
	}

	if (res == -1) {
		if (errno != EINTR) {
#ifdef PHP_WIN32
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "select: %d", WSAGetLastError());
#else
			php_error_docref(NULL TSRMLS_CC, E_WARNING, "select: %s", strerror(errno));
#endif
			return -1;
		}
		return 0;
	}

	maxfd = 0;
	for (ev = TAILQ_FIRST(&EVENT_G(eventqueue)); ev != NULL; ev = next) {
		next = TAILQ_NEXT(ev, ev_next);

		res = 0;
		if (FD_ISSET(ev->ev_fd, st->rfds)) {
			res |= EV_READ;
		}
		if (FD_ISSET(ev->ev_fd, st->wfds)) {
			res |= EV_WRITE;
		}
		if (FD_ISSET(ev->ev_fd, st->efds)) {
			res |= EV_EXCEP;
		}

		res &= ev->ev_events;

		if (res) {
			if ((ev->ev_events & EV_PERSIST) && ev->ev_fd > maxfd) {
				maxfd = ev->ev_fd;
			}
			event_active(ev, res, 1 TSRMLS_CC);
		} else if (ev->ev_fd > maxfd) {
			maxfd = ev->ev_fd;
		}
	}
	st->max_fd = maxfd;

	return 0;
}

const struct php_eventop php_event_select_engine = {
	"select",
	select_event_init,
	select_event_fini,
	select_event_add,
	select_event_del,
	select_event_recalc,
	select_event_dispatch
};


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
